/**
 * \file input-bacnet.c
 * \brief BACnet Monitoring Plugin - input plugin.
 * \copyright Masaryk University, Institute of Computer Science
 * \date 2011
 * \author Pavel CELEDA  <celeda@ics.muni.cz>
 * \author Radek KREJCI  <radek.krejci@mail.muni.cz>
 *
 * Copyright (C) 2011 Masaryk University, Institute of Computer Science
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Masaryk University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: input-bacnet.c 176 2012-01-10 14:55:57Z 98863 $
 */
static const char rcsid[] __attribute__((used)) ="$Id: input-bacnet.c 176 2012-01-10 14:55:57Z 98863 $";

/**
 * \mainpage BACnet Monitoring Plugins
 *
 * \image html doc/doxygen/logo-ics.png
 *
 * \section introduction Introduction
 *
 * The BACnet Monitoring Plugins provide real-time flow data from
 * the Building Automation and Control Networks.
 *
 * \section copyright Copyright Statement
 *
 * Copyright (C) 2011 Masaryk University - Institute of Computer Science. All Rights Reserved.
 *
 */

/**
 * Plugin version
 */
#define VERSION "1.0"

/**
 * Allow processing VLAN headers
 */
#define PROCESS_VLAN
/**
 * Allow processing MPLS headers
 */
#define PROCESS_MPLS

#define _GNU_SOURCE	/**< define getline(), ... GNU extensions */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pcap.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ether.h>
#include <poll.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

//muhahahehe nobody really wants to use it as fast hash, but it is easy to use
#include <openssl/md5.h>

#include <flowmonexp/plugin_input.h>

#include "protocols.h"

/**
 * \defgroup bacnetinput BACnet input plugin
 *
 * \brief Plugin gets packets from a PCAP interface, parses them and prepares
 * flow record structures. Plugin is able to parse standard IP packets as well
 * as specific low level protocols including BACnet (over IP and over Ethernet),
 * ARP, ISMP, LLDP, etc. Also VLAN encapsulation is supported. MPLS
 * encapsulation is currently not supported.
 *
 * By default, plugin reads input data (in PCAP format) from stdin. This can be
 * changed by parameter given to the plugin via flowmonexp command. Parameter
 * can be in a form of path to the input PCAP file or PCAP interface starting
 * with a '@' character. Examples of starting flowmonexp follow.
 *
 * Read data from stdin:
 * \verbatim $ flowmonexp -X input-bacnet.so -I input-bacnet\endverbatim
 * or
 * \verbatim $ flowmonexp -X input-bacnet.so -I input-bacnet:-\endverbatim
 * Read data from local file data.pcap:
 * \verbatim $ flowmonexp -X input-bacnet.so -I input-bacnet:data.pcap\endverbatim
 * Read data from PCAP's interface eth0:
 * \verbatim $ flowmonexp -X input-bacnet.so:@eth0 -I input-bacnet\endverbatim
 *
 * After the input type parameter there can be also specified file with BPF
 * filter for the input data:
 * \verbatim $ flowmonexp -X input-bacnet.so:-,filter.bpf -I input-bacnet\endverbatim
 *
 * Plugin is able to work in a Single Packet Flow mode when each received packet
 * is immediatelly marked as expired flow. This way flowmonexp exports only
 * 1-packet flows. This feature can be enabled by defining macro SPF (e.g.
 * in time of compilation by GCC's -DSPF option). Plugin then uses
 * plugin_input_get_final_flow() function instead of plugin_input_get_flow().
 *
 * @{
 */

#define PCAP_SIZE	1500 	/**< Snapshot length of the PCAP packet */
#define PCAP_TIMEOUT	500 	/**< PCAP read timeout in milliseconds */

/**
 * Set type of the flowmonexp's plugin
 */
unsigned int __attribute__((used))plugin_type = (PLUGIN_TYPE_INPUT);

/**
 * Message printer
 */
#define PRINT(format,args...)
//#define PRINT(format,args...) fprintf(stderr,format,##args)

/**
 * Error message printer
 */
#define PRINTERR(format,args...) fprintf(stderr,format,##args)

/**
 * Warning message printer
 */
#define PRINTWARN(format,args...) fprintf(stderr,format,##args)

/**
 * \brief Plugin's private structure with information passed from init to the
 * calls of main plugin functions.
 */
typedef struct {
	pcap_t *pcap; 		/**< \brief PCAP handler */
	int data_offset; 	/**< \brief BACnet record specific values offset */
	unsigned int counter; 	/**< \brief packet counter */
	int offline; 		/**< \brief The way of reading data (from live interface or from a file) */
} bacnet_input_private_t;

/**
 * \brief BACnet specific part of the flow record
 */
typedef struct {
	uint16_t ethertype; 	/**< \brief flow's EtherType */
	uint8_t control; 	/**< \brief flow's BACnet control value */
	uint8_t hops; 		/**< \brief BACnet NPDU's Hop Count */
	uint8_t msg_type; 	/**< \brief BACnet NPDU's MessageType */
	uint16_t vendor_id; 	/**< \brief BACnet NPDU's Vendor ID */
	uint16_t snet; 		/**< \brief BACnet NPDU's Source Network */
	uint16_t dnet; 		/**< \brief BACnet NPDU's Destination Network */
	uint8_t slen; 		/**< \brief BACnet NPDU's source address length */
	uint8_t dlen; 		/**< \brief BACnet NPDU's source address length */
	uint8_t sadr[7]; 	/**< \brief BACnet NPDU's source address */
	uint8_t dadr[7]; 	/**< \brief BACnet NPDU's destination address */
} bacnet_record_t;

/**
 * \brief Parse Ethernet header and fill required information into the flow
 * record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_eth (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr);

/**
 * \brief Parse Logical Link Control header and fill required information into the
 * flow record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_llc (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr);

/**
 * \brief Parse BACnet Virtual Link Control header and fill required information
 * into the flow record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet_vlc (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr);

/**
 * \brief Parse IP header and fill required information into the flow
 * record
 *
 * \param[in] ethtype Packet's EtherType value.
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet_ip (uint16_t ethtype, unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr);

/**
 * \brief Parse BACnet headers and fill required information into the flow
 * record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr);

/**
 * Plugin description structure for internal use in flowmonexp
 */
static plugin_desc_t plugin_desc = {
	"input-bacnet", 		/**< Plugin name */
	"Version " VERSION
	"\nflowmonexp input plugin processing BACnet traffic\
	\nOne of input parameters:\
	\n\t-     read data from stdin\
	\n\tfile  read data from pcap file\
	\n\t@if   read data from pcap interface if\
	\nSeparated by ',' there can be specified filename containing BPF filter(s)", /**< Plugin help string */
	sizeof(bacnet_record_t), 	/**< Requested data size after standard flow record (MUST BE CONSTANT) */
	0 				/**< Plugin must have packet input, valid only for process plugin */
};

/**
 * \brief API for flowmonexp to get plugin description structure.
 * \return Plugin description structure
 */
plugin_desc_t *plugin_input_desc()
{
	return (&plugin_desc);
}

/**
 * \brief Validity checker for BACnet record's control field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_control(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (pr->ethertype == BACNET_ETH_TYPE || pr->ethertype == BACNET_IP_TYPE) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief General validity checker that always fails
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_fail(void *self, flow_record_t *r)
{
	return (0);
}

/**
 * \brief Validity checker for BACnet record's ethertype field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_ethertype(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (pr->ethertype != 0) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Validity checker for BACnet record's hop count and dnet fields
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_dst(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (pr->control & BACNET_CTRL_DST_MASK) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Validity checker for BACnet record's hop dadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_dstadr(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (validity_dst (self, r) && pr->dlen > 0) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Validity checker for BACnet record's snet field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_src(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (pr->control & BACNET_CTRL_SRC_MASK) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Validity checker for BACnet record's sadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_srcadr(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (validity_src (self, r) && pr->slen > 0) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Validity checker for BACnet record's msg_type field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int validity_msg_type(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	if (pr->control & BACNET_CTRL_MSGTYPE_MASK) {
		return (1);
	} else {
		return (0);
	}
}

/**
 * \brief Get actual length for BACnet record's 8bit static fields
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int value_length_8(void *self, flow_record_t *r)
{
	return (sizeof(uint8_t));
}

/**
 * \brief Get actual length for BACnet record's 16bit static fields
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int value_length_16(void *self, flow_record_t *r)
{
	return (sizeof(uint16_t));
}

/**
 * \brief Get actual length of the BACnet record's dadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int value_length_dadr(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	return (pr->dlen);
}

/**
 * \brief Get actual length of the BACnet record's sadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] r Standard flowmonexp's flow record
 * \return
 *   - zero as invalid
 *   - nonzero as valid
 */
int value_length_sadr(void *self, flow_record_t *r)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(r,p->data_offset);
	return (pr->slen);
}

/**
 * \brief Get value of the BACnet record's ethertype field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_ethertype(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->ethertype), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's control field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_control(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->control), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's hop_count field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_hops(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->hops), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's msg_type field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_msg_type(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->msg_type), 0, sizeof(uint8_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's vendor_id field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_vendor_id(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->vendor_id), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's dnet field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_dnet(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->dnet), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's dadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_dadr(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->dadr[0]), 0, pr->dlen, len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's snet field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_snet(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->snet), 0, sizeof(uint16_t), len, to_network_byte_order);
}

/**
 * \brief Get value of the BACnet record's sadr field
 * \param[in] self Plugin's specific record part (bacnet_record_t) allocated after standard flow record
 * \param[in] record Standard flowmonexp's flow record
 * \param[in] dst Where to copy resulted data
 * \param[in] len Data length
 * \param[in] to_network_byte_order if other than zero, convert value to network byte order (work only for lengths: 1,2,4,8)
 */
void value_fill_sadr(void *self, flow_record_t *record, void *dst, int len, int to_network_byte_order)
{
	bacnet_input_private_t *p = (bacnet_input_private_t*) self;
	bacnet_record_t *pr = PLUGIN_DATA(record,p->data_offset);
	record_universal_get (dst, &(pr->sadr[0]), 0, pr->slen, len, to_network_byte_order);
}

/**
 * \brief Init flowmonexp's getters for BACnet record's specific fields
 *
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \param[in,out] getter_list List of available getters for a complete flow record
 */
void plugin_input_getter_init(void *plugin_private, flow_record_getter_t **getter_list)
{
	PRINT("getter init started\n");
	getter_add (getter_list, "ETHERTYPE", sizeof(uint16_t), plugin_private, &validity_ethertype, &value_length_16, &value_fill_ethertype);
	getter_add (getter_list, "BACNET_CONTROL", sizeof(uint8_t), plugin_private, &validity_control, &value_length_8, &value_fill_control);
	getter_add (getter_list, "BACNET_HOP_COUNT", sizeof(uint8_t), plugin_private, &validity_dst, &value_length_8, &value_fill_hops);
	getter_add (getter_list, "BACNET_MESSAGE_TYPE", sizeof(uint8_t), plugin_private, &validity_msg_type, &value_length_8, &value_fill_msg_type);
	getter_add (getter_list, "BACNET_VENDOR_ID", sizeof(uint16_t), plugin_private, &validity_fail, &value_length_16, &value_fill_vendor_id);
	getter_add (getter_list, "BACNET_SNET", sizeof(uint16_t), plugin_private, &validity_src, &value_length_16, &value_fill_snet);
	getter_add (getter_list, "BACNET_DNET", sizeof(uint16_t), plugin_private, &validity_dst, &value_length_16, &value_fill_dnet);
	getter_add (getter_list, "BACNET_SADR", -1, plugin_private, &validity_srcadr, &value_length_sadr, &value_fill_sadr);
	getter_add (getter_list, "BACNET_DADR", -1, plugin_private, &validity_dstadr, &value_length_dadr, &value_fill_dadr);
	PRINT("getter init finished\n");
}

/**
 * \brief Initialize all needed structures/devices/files etc. The function is
 * called once as the first part of the plugin.
 *
 * \param[in] params Plugin parameters as string from command line
 * \param[in] full_packet Specify if plugin_input_get_packet() function (if implemented) is forced to return full packet or some cropping can be done
 * \param[in] data_offset Offset of the plugin specific data (bacnet_record_t) in flowmonexp's standard flow record
 * \return Created and initialized plugin internal structure
 *         (bacnet_eth_private_t)
 */
void *plugin_input_init(char *params, int full_packet, int data_offset)
{
	bacnet_input_private_t *retval;
	char errbuf[PCAP_ERRBUF_SIZE];
	char *filename, *next_param = NULL, *line = NULL;
	struct bpf_program filter;
	FILE *filter_file;
	size_t size;

	/* Prepare return structure */
	retval = malloc (sizeof(bacnet_input_private_t));
	if (!retval) {
		return (NULL);
	}

	retval->offline = 1;
	if (params != NULL) {
		next_param = strchr (params, ',');
		if (next_param != NULL) {
			next_param[0] = 0;
			next_param++;
		}
	}

	/* check given parameter */
	if (!params || params[0] == '-' || params[0] == 0) {
		/* stdin */
		filename = strdup ("-");
	} else {
		if (params[0] == '@') {
			/* live device */
			retval->offline = 0;
			filename = strdup (&params[1]);
		} else {
			/* offline file */
			retval->offline = 1;
			filename = strdup (params);
		}
	}

	/* open a saved PCAP capture file or device */
	if (retval->offline == 1) {
		PRINT ("BACnet plugin opens file %s\n", filename);
		retval->pcap = pcap_open_offline (filename, errbuf);
	} else {
		PRINT ("BACnet plugin opens live device %s\n", filename);
		retval->pcap = pcap_open_live (filename, PCAP_SIZE, 1, 0, errbuf);
	}

	free (filename);
	if (retval->pcap == NULL) {
		PRINTERR ( "Unable to init pcap: %s\n", errbuf);
		free (retval);
		return (NULL);
	}

	/* set BPF filters */
	while (next_param != NULL) {
		PRINT ("setting BPF ");
		params = next_param;
		next_param = strchr (params, ',');
		if (next_param != NULL) {
			next_param[0] = 0;
			next_param++;
		}

		PRINT ("from file %s\n", params);
		if ((filter_file = fopen(params, "r")) == NULL) {
			PRINTERR ("Setting BPF filter failed: %s\n", strerror (errno));
			free (retval);
			return (NULL);
		}
		line = NULL;
		while (getline (&line, &size, filter_file) > 0) {
			PRINT ("Filtering with BPF: %s\n", line);
			if (pcap_compile (retval->pcap, &filter, line, 1, 0) != 0) {
				PRINTERR ("Compiling BPF filter failed: %s\n", pcap_geterr (retval->pcap));
				free (retval);
				return (NULL);
			}
			if (pcap_setfilter (retval->pcap, &filter)) {
				PRINTERR ("Setting BPF filter failed: %s\n", pcap_geterr (retval->pcap));
				free (retval);
				return (NULL);
			}
			free (line);
			line = NULL;
		}

		fclose (filter_file);
	}
	retval->data_offset = data_offset;
	retval->counter = 0;
	return (retval);
}

static MD5_CTX md5;

/**
 * \fn uint64_t plugin_input_get_final_flow(void *plugin_private, flow_record_t * record)
 * \brief Get flow record implementation - returned flow is immediatelly marked as expired
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \param[in] record Flow record from input plugin(s)
 * \return
 *   - zero on success,
 *   - nonzero on error
 */
/**
 * \fn uint64_t plugin_input_get_flow(void *plugin_private, flow_record_t * record)
 * \brief Get flow record implementation - add flow record into the flowmonexp's flow cache
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \param[in,out] record Flow record from input plugin(s)
 * \return
 *   - zero on error,
 *   - hash of the flow record (actually of the flow key items)
 */
#ifdef DOXYGEN /* This part is ONLY for generating Doxygen documentation!!! */
uint64_t plugin_input_get_final_flow(void *plugin_private, flow_record_t * record);
uint64_t plugin_input_get_flow(void *plugin_private, flow_record_t * record);
#endif

#ifdef SPF
/* Single packet flow implementation - each flow contain exactly 1 packet */
#ifndef DOXYGEN
#define GET_FAILED 1 /**< error return value of the plugin_input_get_final_flow() */
#endif
uint64_t plugin_input_get_final_flow(void *plugin_private, flow_record_t * record)
{
#else
#ifndef DOXYGEN
#define GET_FAILED 0 /**< \internal error return value of the plugin_input_get_flow() */
#endif
uint64_t plugin_input_get_flow(void *plugin_private, flow_record_t * record)
{
#endif
	bacnet_input_private_t *p = (bacnet_input_private_t*) plugin_private;
	bacnet_record_t *precord = PLUGIN_DATA(record,p->data_offset);
	struct pcap_pkthdr hdr;
	unsigned char *buf;
	int datalink;
	uint64_t hash[2];
	int ret;

	memset ((void*) precord, 0, sizeof(bacnet_record_t));

	PRINT("packet No: %d\n", ++(p->counter));

	/* read the data */
	buf = (unsigned char*) pcap_next (p->pcap, &hdr);

	/* check read buffer */
	if ((p->offline == 1) && (buf == NULL)) {
		/* when we read files, NULL always means EOF and we want to stop exporter */
		sleep (5); /* let the output plugins do what they need - the time should be probably (in)active timeout? */
		raise (SIGQUIT);
		sleep (5);
		return GET_FAILED;
	}

	/* set time into flow record */
	record->flow_end = record->flow_start = ((uint64_t) hdr.ts.tv_sec * 1000 * 1000 * 1000) + ((uint64_t) hdr.ts.tv_usec * 1000);
	/* set packet count to 1 */
	record->flow_packets = 1;
	/* set real octets length (packet may be stripped from pcap) */
	record->flow_octets = hdr.len;

	/* begin parse packet */
	datalink = pcap_datalink (p->pcap);

	MD5_Init (&md5);
	if (datalink == DLT_EN10MB) {
		/* parse Ethernet header */
		ret = parse_eth (buf, (uint32_t) hdr.caplen, record, precord);
		PRINT("parse_eth returned %d\n", ret);
		if (ret != EXIT_SUCCESS) {
			return GET_FAILED;
		} else {
#ifdef SPF
			return 0;
#else
			//final md5 transform and final work
			MD5_Final ((unsigned char *) hash, &md5);
			return ((hash[0] ^ hash[1]) | FLOW_OK_BIT);
#endif
		}
	}

	return GET_FAILED;
}

/**
 * \brief Parse Ethernet header and fill required information into the flow
 * record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_eth (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr)
{
	uint16_t ethtype;
	uint32_t orig_caplen = caplen;
	uint64_t internal_hash = 0;
	struct ethhdr *eth;
	unsigned char* next_hdr = buf;
	int ret = EXIT_SUCCESS;
	vlan_t *vlan;
	l2_t *mac;
	uint32_t bytes = record->flow_octets; /* store it for later use */

	if (caplen < ETH_HLEN
	)
		return (EXIT_FAILURE);

	eth = (struct ethhdr*) buf;
	ethtype = ntohs (eth->h_proto);

	if (ethtype <= ETH_DATA_LEN) {
		/* EtherType is actually a size of user data, so after Ethernet header
		 * follows the LLC header */
		PRINT("LLC instead of 802.3\n");
		return (parse_llc (next_hdr + ETH_HLEN, caplen - ETH_HLEN, record, pr));
	} else {
		PRINT("802.3\n");

		/* get MAC addresses */
		if ((mac = record_mac(record)) != NULL) {
			memcpy (mac->src_mac, eth->h_source, ETH_ALEN);
			memcpy (mac->dst_mac, eth->h_dest, ETH_ALEN);
			MD5_Update (&md5, mac, 2 * ETH_ALEN);
		}

		/* move from simple 802.3 header to PDU */
		next_hdr += ETH_HLEN;
		caplen -= ETH_HLEN;
	}

	recurse:
	/* decide what to do now */
	switch (ethtype) {
	case ETH_P_IPV6: /* IPv6 */
	case ETH_P_IP: /* IPv4 */
		PRINT("IP, %p\n", buf);
		/* includes ICMP as well */
		pr->ethertype = ethtype;
		if ((internal_hash = record_process_packet (buf, (int) orig_caplen, record)) == 0) {
			return (EXIT_FAILURE);
		}
		/* parse packet as BACnet over IP? */
		if ((record->l4.protocol == IPPROTO_UDP) && (record->l4.src_port == BACNET_UDP_PORT) && (record->l4.dst_port == BACNET_UDP_PORT)) {
			/* BACnet over IP */
			parse_bacnet_ip (ethtype, next_hdr, caplen, record, pr);
		}
		MD5_Update (&md5, &internal_hash, 8);

		/* WARNING - TODO */
		/* here we store size of the ethernet frame even if this is an IP packet
		 * and field BYTES should normally contain size of the IP packet. This
		 * unifies the value for all types of traffic
		 */
		record->flow_octets = bytes;

		break;
	case ETH_P_8021Q: /* VLAN */
	case _ETH_P_8021QINQ: /* Stacked VLAN, aka QinQ */
	case _ETH_P_8021QINQ_DOUBLE: /* double stacked */
	case _ETH_P_8021QINQ_TRIPPLE: /* tripple stacked */
	case _ETH_P_8021AD: /* 802.1ad */
		PRINT("VLAN\n");
		if (caplen < VLAN_HLEN) {
			return (EXIT_FAILURE);
		}

		/* store current VLAN ID if there is space in record */
		if ((vlan = record_vlan (record)) != NULL) {
			if (vlan->count < VLAN_MAX) {
				PRINT("VLAN count: %d\n", vlan->count);
				vlan->id[vlan->count] = htons((*(uint16_t *) next_hdr) & VLAN_ID_MASK);
				MD5_Update (&md5, &(vlan->id[vlan->count]), 2);
				vlan->count++;
			}
		}

		/*
		 * Ethertype of following header is stored in higher 2B
		 * of VLAN header (sometimes it is described as Ethertype
		 * from Ethernet header moved after VLAN header that
		 * contains its own Ethertype in lower 2B)
		 */
		ethtype = ntohs (*(uint16_t *) (next_hdr + 2));
		next_hdr += VLAN_HLEN;
		caplen -= VLAN_HLEN;

		/* repeat parsing after VLAN header */
		goto recurse;
	case ETH_P_MPLS_UC: /* MPLS */
	case ETH_P_MPLS_MC:
		PRINT("MPLS\n");
		pr->ethertype = ethtype;
		ret = EXIT_SUCCESS;
		//TODO? ret = parse_mpls(next_hdr, caplen, record, md5);
		break;
	default: /* Other protocols */
		if (ethtype <= ETH_DATA_LEN) {
			/* There is LLC header */
			ret = parse_llc (next_hdr, caplen, record, pr);
		} else {
			/* something general - store its EtherType and finnish packet processing */
			pr->ethertype = ethtype;
			ret = EXIT_SUCCESS;
		}
		break;
	}

	return (ret);

}

/**
 * \brief Parse IP header and fill required information into the flow
 * record
 *
 * \param[in] ethtype Packet's EtherType value.
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet_ip (uint16_t ethtype, unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr)
{
	struct ip *ip;
	struct ip6_hdr *ip6;
	unsigned char* bacnet_vlc = buf;
	int ret = 1;

	switch (ethtype) {
	case ETH_P_IPV6: /* IPv6 */
		PRINT ("BACnet over IPv6\n");
		ip6 = (struct ip6_hdr*) buf;
		/* TODO */
		break;
	case ETH_P_IP: /* IPv4 */
		ip = (struct ip*) buf;
		bacnet_vlc += ((ip->ip_hl * 4) + UDP_HDR_SIZE);
		caplen -= ((ip->ip_hl * 4) + UDP_HDR_SIZE);
		PRINT ("BACnet over IPv4, %p (%d)\n", buf, ip->ip_hl * 4);
		/* go into BACnet Virtual Link Control */
		ret = parse_bacnet_vlc (bacnet_vlc, caplen, record, pr);
		break;
	default:
		return 1;
		break;
	}

	return ret;
}

/**
 * \brief Parse BACnet Virtual Link Control header and fill required information
 * into the flow record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet_vlc (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr)
{
	int bvlc_length;
	uint8_t bvlc_function = buf[1];

	PRINT ("BVLC, %p\n", buf);
	pr->ethertype = BACNET_IP_TYPE;

	if (bvlc_function > 0x08) {
		/*  We have a constant header length of BVLC of 4 in every
		 *  BVLC-packet forewarding an NPDU. Beware: Changes in the
		 *  BACnet-IP-standard may break this.
		 *  At the moment, no functions above 0x0b
		 *  exist (Addendum 135a to ANSI/ASHRAE 135-1995 - BACnet)
		 */
		bvlc_length = 4;
	} else if (bvlc_function == 0x04) {
		/* 4 Bytes + 6 Bytes for B/IP Address of Originating Device */
		bvlc_length = 10;
	} else {
		/*  BVLC-packets with function below 0x09 contain
		 *  routing-level data (e.g. Broadcast Distribution)
		 *  but no NPDU for BACnet, so bvlc_length goes up to the end
		 *  of the captured frame.
		 */
		return (EXIT_SUCCESS);
	}
	return (parse_bacnet (buf + bvlc_length, caplen - bvlc_length, record, pr));
}

/**
 * \brief Parse Logical Link Control header and fill required information into the
 * flow record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_llc (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr)
{
	int hlen = 0;

	/* check if LLC contains BACnet */
	if ((buf[0] & LLC_SAP_MASK) != LLC_SAP_BACNET) {
		PRINT ("Unsupported LLC\n");
		return (EXIT_FAILURE);
	}

	/*
	 * get LLC header length according to lowest 2 bits in control field:
	 * 0x03 - U-format PDUs witn an 8-bit control field
	 * 0x01 - S-format PDUs, with a 16-bit control field
	 * 0x00,0x02 - I-format PDUs, with a 16-bit control and sequence numbering field
	 */
	hlen = ((buf[2] & 0x03) == 0x03) ? 3 : 4;
	PRINT ("LLC type %x\n",buf[2] & 0x03);

	/* parse BACnet inside LLC */
	pr->ethertype = (BACNET_ETH_TYPE);
	//MD5_Update(&md5,&(pr->ethertype),2);
	return (parse_bacnet (buf + hlen, caplen - hlen, record, pr));
}

/**
 * \brief Parse BACnet headers and fill required information into the flow
 * record
 *
 * \param[in] buf Data buffer.
 * \param[in] caplen Length of portion present.
 * \param[in,out] record Flow record structure.
 * \param[in] pr Plugin's private structure with BACnet specific record offset
 * \retval 0 on success, nonzero otherwise.
 */
int parse_bacnet (unsigned char* buf, uint32_t caplen, flow_record_t *record, bacnet_record_t *pr)
{
	PRINT("BACnet, %p\n", buf);
	uint8_t *data;
	int i, len;

	struct bacnet_npdu *npdu = (struct bacnet_npdu*) buf;
	PRINT("\tVersion: %d\n", npdu->version);

	PRINT("\tControl: %x\n", npdu->control);
	pr->control = npdu->control;

	data = &(npdu->data[0]);

	if (npdu->control & BACNET_CTRL_DST_MASK) {
		PRINT("\tDNET: %d\n", ntohs(*(uint16_t*)&data[0]));
		pr->dnet = ntohs (*(uint16_t*) &data[0]);
		MD5_Update (&md5, &(pr->dnet), sizeof(uint16_t));

		PRINT("\tDLEN: %d\n", data[2]);
		pr->dlen = data[2];
		len = 3;
		if (data[2] > 0) {
			PRINT ("\tDADR: ");
			for (i = 0; i < data[2]; i++) {
				PRINT("%02X",data[3+i]);
				len++;
			}PRINT ("\n");
			memcpy (&(pr->dadr[0]), &(data[3]), data[2]);
			MD5_Update (&md5, &(pr->dadr), pr->dlen);
		}
		data = &(data[len]);
	}

	if (npdu->control & BACNET_CTRL_SRC_MASK) {
		PRINT("\tSNET: %d\n", ntohs(*(uint16_t*)&data[0]));
		pr->snet = ntohs (*(uint16_t*) &data[0]);
		MD5_Update (&md5, &(pr->snet), sizeof(uint16_t));

		PRINT("\tSLEN: %d\n", data[2]);
		pr->slen = data[2];
		len = 3;
		if (data[2] > 0) {
			PRINT ("\tSADR: ");
			for (i = 0; i < data[2]; i++) {
				PRINT("%02X",data[3+i]);
				len++;
			}PRINT ("\n");
			memcpy (&(pr->sadr[0]), &(data[3]), data[2]);
			PRINT ("\tSADR: ");
			for (i = 0; i < pr->slen; i++) {
				PRINT("%02X",pr->sadr[i]);
			}PRINT ("\n");
			MD5_Update (&md5, &(pr->sadr), pr->slen);
		}
		data = &(data[len]);
	}

	if (npdu->control & BACNET_CTRL_DST_MASK) {
		PRINT ("\tHOP COUNT: %d\n", data[0]);
		pr->hops = data[0];
		data = &(data[1]);
	}

	/* message type appears last in BACnet header */
	if (npdu->control & BACNET_CTRL_MSGTYPE_MASK) {
		PRINT ("\n\tMSG TYPE: %d\n", data[0]);
		pr->msg_type = data[0];
		data = &(data[1]);
	}

	return (EXIT_SUCCESS);
}

/**@}*/

