/**
 * \file protocols.h
 * \brief BACnet Monitoring Plugin - addon for standard network headers.
 * \copyright Masaryk University, Institute of Computer Science
 * \date 2011
 * \author Pavel CELEDA  <celeda@ics.muni.cz>
 * \author Radek KREJCI  <radek.krejci@mail.muni.cz>
 *
 * Copyright (C) 2011 Masaryk University, Institute of Computer Science
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Masaryk University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: protocols.h 166 2011-12-07 13:31:04Z 206086 $
 */

#ifndef _PROTOCOLS_H_
#define _PROTOCOLS_H_

#include <netinet/if_ether.h>

/*
 * Ethernet
 */
/* some possibly missing ethertypes */
#ifndef ETH_P_8021Q
#define ETH_P_8021Q	0x8100		/**< 802.1Q VLAN Extended Header  */
#endif

#define ETH_P_ISMP	0x81FD		/**< InterSwitch Message Protocol */

#ifndef ETH_P_MPLS_UC
#define ETH_P_MPLS_UC	0x8847		/**< MPLS Unicast traffic		*/
#endif

#ifndef ETH_P_MPLS_MC
#define ETH_P_MPLS_MC	0x8848		/**< MPLS Multicast traffic	*/
#endif

#define ETH_P_LLDP	0x88CC		/**< Link Layer Discovery Protocol */

#define _ETH_P_8021AD   0x88A8          /**< 802.1ad (obsoletes QinQ) */
#define _ETH_P_8021QINQ 0x9100          /**< 802.1QinQ (VLAN in VLAN) - CISCO */
#define _ETH_P_8021QINQ_DOUBLE  0x9200  /**< 802.1QinQ (2 VLANs in VLAN) */
#define _ETH_P_8021QINQ_TRIPPLE 0x9300  /**< 802.1QinQ (3 VLANs in VLAN) */

/*
 * LLC (Logical Link Control)
 */
#define LLC_SAP_MASK	0xFE 	/**< Mask of the LLC's Service Access Point value in LLC header */
#define LLC_SAP_BACNET	0x82 	/**< LLC's Service Access Point value for BACnet protocol */

/*
 * VLAN
 */
#define VLAN_HLEN	4 	/**< Length of the VLAN header */
#define VLAN_ID_MASK	0xFF0F  /**< Mask (LE) of the VLAN ID in the header */

/*
 * MPLS
 */
#define MPLS_HLEN 4 		/**< Length of the MPLS header */

#define MPLS_LABEL_MASK	0xfffff000 	/**< Mask of the label value in the MPLS header */
#define MPLS_LABEL_SHIFT    12 		/**< Number of bits to shift label value */
#define	MPLS_EXP_MASK	0x00000e00 	/**< Mask of the EXP (experimental functions) value in the MPLS header */
#define MPLS_EXP_SHIFT  9 		/**< Number of bits to shift EXP value */
#define	MPLS_STACK_MASK	0x00000100 	/**< Mask of the STACK (stack functions) value in the MPLS header */
#define MPLS_STACK_SHIFT    8 		/**< Number of bits to shift STACK value */
#define MPLS_TTL_MASK	0x000000ff 	/**< Mask of the TTL (Time-to-Live) value in the MPLS header */
#define MPLS_TTL_SHIFT  0 		/**< Number of bits to shift TTL value */

/*
 * UDP
 */
#define UDP_HDR_SIZE 8 		/**< Length of the UDP header */


/**
 * \brief BACnet NPDU structure
 */
struct __attribute__ ((packed)) bacnet_npdu {
	/**
	 * \brief Version field
	 */
	uint8_t version;
	/**
	 * \brief Control field value that specifies a content of the rest of the NPDU
	 */
	uint8_t control;
	/**
	 * \brief Specific NPDU values according to control value
	 */
	uint8_t data[];
};

#define BACNET_CTRL_MSGTYPE_MASK 0x80 	/**< Mask for MessageType enable bit of Control value in BACnet NPDU */
#define BACNET_CTRL_DST_MASK 0x20 	/**< Mask for Destination enable bit of Control value in BACnet NPDU */
#define BACNET_CTRL_SRC_MASK 0x08 	/**< Mask for Source enable bit of Control value in BACnet NPDU */
#define BACNET_UDP_PORT 0xbac0 		/**< BACnet's standard UDP port */

#define BACNET_IP_TYPE 0x81 		/**< BACnet over IP mapping into the EtherType value (This value of EtherType is actually invalid because it means a size of the packet in real traffic) */
#define BACNET_ETH_TYPE LLC_SAP_BACNET 	/**< BACnet over Ethernet mapping into the EtherType value (This value of EtherType is actually invalid because it means a size of the packet in real traffic) */

#endif /* _PROTOCOLS_H_ */

