/**
 * \file export-bacnet-csv.c
 * \brief BACnet Monitoring Plugin - CSV output plugin.
 * \copyright Masaryk University, Institute of Computer Science
 * \date 2011
 * \author Pavel CELEDA  <celeda@ics.muni.cz>
 * \author Radek KREJCI  <radek.krejci@mail.muni.cz>
 *
 * Copyright (C) 2011 Masaryk University, Institute of Computer Science
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Masaryk University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: export-bacnet-csv.c 175 2012-01-10 12:26:31Z 98863 $
 */
static const char rcsid[] __attribute__((used)) ="$Id: export-bacnet-csv.c 175 2012-01-10 12:26:31Z 98863 $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <flowmonexp/plugin_export.h>

/**
 * \defgroup bacnetcsv BACnet CSV output plugin
 *
 * \brief Plugin prints commands for PostgreSQL to insert exported data into
 * the database. Plugin supports optional dividing of the exported data into
 * the database partitions according to a date (for each
 * year/month/day/hour/minute).
 *
 * Example of using the plugin with flowmonexp and enabled partitioning for each
 * day:
 * \verbatim $ flowmonexp -X export-bacnet-csv.so -I pcap:device=eth0 -E export-bacnet-csv:day\endverbatim
 *
 * @{
 */

/**
 * Plugin version
 */
#define VERSION "1.0"

/**
 * Set type of the flowmonexp's plugin
 */
unsigned int __attribute__((used))plugin_type = (PLUGIN_TYPE_EXPORT);

/**
 * Message printer
 */
//#define PRINT(format,args...) fprintf(stderr,format,##args)
#define PRINT(format,args...)

/**
 * Error message printer
 */
#define PRINTERR(format,args...) fprintf(stderr,format,##args)

/**
 * Warning message printer
 */
#define PRINTWARN(format,args...) fprintf(stderr,format,##args)

/**
 * SQL printer init
 */
#define PRINT_SQL_INIT char obuf[16384],dummy[1024];obuf[0]='\0'
//#define PRINT_SQL_INIT
/**
 * SQL commands printer
 */
#define PRINT_SQL(format,args...) sprintf(dummy,format,##args);strcat(obuf, dummy)
//#define PRINT_SQL(format,args...) printf(format,##args)
/**
 * SQL printer flush
 */
#define PRINT_SQL_FLUSH printf("%s", obuf);fflush(stdout)
//#define PRINT_SQL_FLUSH fflush(stdout)

/**
 * Size of the buffer for printing network address
 */
#define ADDR_BUF_SIZE 128

/**
 * \brief Time units for DB partitioning
 */
enum TIME_UNITS {
	TU_NO, 		/**< No partitioning */
	TU_YEAR,	/**< New partition is available each year */
	TU_MONTH, 	/**< New partition is available each month */
	TU_DAY, 	/**< New partition is available each day */
	TU_HOUR, 	/**< New partition is available each hour */
	TU_MINUTE 	/**< New partition is available each minute */
};

/**
 * \brief Plugin's private structure with information passed from init to the calls of
 * PLUGIN_EXPORT_* functions.
 */
typedef struct{
	int counter;                   /**< \brief number of exported flows */
	enum TIME_UNITS partition;     /**< \brief time units for DB partitioning */
	flow_record_getter_t *getters; /**< \brief list of specified getters for output */
} bacnet_export_csv_private_t;

/**
 * \brief Plugin description structure for internal use in flowmonexp
 */
static plugin_desc_t pcap_desc = {
	"export-bacnet-csv", 	/**< Plugin name */
	"Version " VERSION
	"\nPlugin which outputs all records to stdout in csv format. A time unit can be set as parameter for output partitioning - allowed values include year, month, day, hour and minute.", /**< plugin help */
	0, 	/**< Requested data size after standard flow record (MUST BE CONSTANT) */
	0 	/**< Plugin must have packet input, valid only for process plugin */
};

/**
 * \brief List of possible getters that have implemented a printer in the EXPORT
 * function. Feel free to add any new item with its printer implemented in
 * PLUGIN_EXPORT_EXPORT, but the last item must not be changed. This list do not
 * actually affects currently printed items - see selected_getters[].
 */
enum GETTER_ID {
	FLOW_START_USEC,
	FLOW_END_USEC,
	L2_VLAN_0,
	L2_VLAN_1,
	L2_VLAN_2,
	L2_VLAN_3,
	L2_SRC_MAC,
	L2_DST_MAC,
	ETHERTYPE,
	L3_PROTO,
	L3_IPV4_ADDR_SRC,
	L3_IPV4_ADDR_DST,
	L3_IPV4_TOS,
	L3_IPV6_ADDR_SRC,
	L3_IPV6_ADDR_DST,
	L4_PROTO,
	L4_TCP_FLAGS,
	L4_PORT_SRC,
	L4_PORT_DST,
	L4_ICMP_TYPE_CODE,
	BACNET_CONTROL,
	BACNET_HOP_COUNT,
	BACNET_MESSAGE_TYPE,
	BACNET_VENDOR_ID,
	BACNET_SNET,
	BACNET_DNET,
	BACNET_SADR,
	BACNET_DADR,
	PACKETS,
	BYTES,
	/* This MUST be always the last one */
	GETTER_MAX
};

/**
 * \brief Auxiliary container connecting name string of the getter with the ID needed
 * for processing in PLUGIN_EXPORT_EXPORT function.
 */
struct getter_pair {
	char *name;        /**< \brief name of the getter */
	enum GETTER_ID id; /**< \brief ID of the getter assigned via enum */
};

/**
 * \brief The list of the printed getters. The order in this list determine the order
 * of the information printed to the output.
 *
 * The case branch in PLUGIN_EXPORT_EXPORT's switch of the first item should
 * contain start of the printed string - in our case there is INSERT... SQL
 * command.
 */
struct getter_pair selected_getters[] = {
	{"FLOW_START_USEC", FLOW_START_USEC},
	{"FLOW_END_USEC", FLOW_END_USEC},
	{"L2_VLAN_0", L2_VLAN_0},
	{"L2_SRC_MAC", L2_SRC_MAC},
	{"L2_DST_MAC", L2_DST_MAC},
	{"ETHERTYPE", ETHERTYPE},
	{"L3_PROTO", L3_PROTO},
	{"L3_IPV4_ADDR_SRC", L3_IPV4_ADDR_SRC},
	{"L3_IPV4_ADDR_DST", L3_IPV4_ADDR_DST},
	{"L3_IPV6_ADDR_SRC", L3_IPV6_ADDR_SRC},
	{"L3_IPV6_ADDR_DST", L3_IPV6_ADDR_DST},
	{"L4_PROTO", L4_PROTO},
	{"L4_TCP_FLAGS", L4_TCP_FLAGS},
	{"L4_PORT_SRC", L4_PORT_SRC},
	{"L4_PORT_DST", L4_PORT_DST},
	{"L4_ICMP_TYPE_CODE", L4_ICMP_TYPE_CODE},
	{"BACNET_CONTROL", BACNET_CONTROL},
	{"BACNET_HOP_COUNT", BACNET_HOP_COUNT},
	{"BACNET_MESSAGE_TYPE", BACNET_MESSAGE_TYPE},
	{"BACNET_SNET", BACNET_SNET},
	{"BACNET_DNET", BACNET_DNET},
	{"BACNET_SADR", BACNET_SADR},
	{"BACNET_DADR", BACNET_DADR},
	{"PACKETS", PACKETS},
	{"BYTES", BYTES},
	/* This MUST be always the last item */
	{NULL, GETTER_MAX}
};

/*
 * Flowmonexp API implementations
 */

/**
 * \brief API for flowmonexp to get plugin description structure.
 * \return Plugin description structure
 */
plugin_desc_t *plugin_export_desc()
{
	return(&pcap_desc);
}

/**
 * \brief Initialize all needed structures/devices/files etc. The function is
 * called once as the first part of the plugin.
 *
 * \param[in] params Plugin parameters as string from command line
 * \param[in] getter_list List of available getters for a complete flow record
 * \return Created and initialized plugin internal structure
 *         (export_bacnet_csv_private_t)
 */
void *plugin_export_init(char *params, flow_record_getter_t *getter_list)
{
	bacnet_export_csv_private_t *retval;
	flow_record_getter_t *tmp = getter_list;
	int i;

	PRINT("CSV Output plugin init start\n");
	retval = malloc(sizeof(bacnet_export_csv_private_t));
	if(!retval){
		return(NULL);
	}

	/* init internal plugin structure */
	retval->counter = 0;
	retval->partition = TU_NO;
	retval->getters = NULL;

	/* Handle settings for db partitioning */
	if (params != NULL) {
		if (strlen (params) == 0) {
			PRINTERR ( "Invalid parameters\n");
			exit (-1);
		} else if (!strcmp ("year", params)) {
			retval->partition = TU_YEAR;
		} else if (!strcmp ("month", params)) {
			retval->partition = TU_MONTH;
		} else if (!strcmp ("day", params)) {
			retval->partition = TU_DAY;
		} else if (!strcmp ("hour", params)) {
			retval->partition = TU_HOUR;
		} else if (!strcmp ("minute", params)) {
			retval->partition = TU_MINUTE;
		} else {
			PRINTWARN ( "Unknown partitioning parameter (%s) - no partitioning used\n", params);
		}
	}

	/* get selected getters */
	i = 0;
	while (selected_getters[i].id != GETTER_MAX) {
		tmp = getter_by_name (getter_list, selected_getters[i].name);
		if (tmp != NULL) {
			PRINT("getter: %s - %i\n",tmp->name, tmp->length);
			getter_copy_to(&(retval->getters),tmp);
		}
		i++;
	}

	PRINT("CSV Output plugin init end\n");

	/* enable stdout line buffering. */
	setvbuf(stdout, (char *)NULL, _IOLBF, 0);

	return(retval);
}

/**
 * \brief Print titles of used columns in records.
 *
 * \param[in] getters List of available getters for a complete flow record
 *
 */
void print_headers(flow_record_getter_t *getters)
{
	while(getters && getters->name){
		printf ("%s,", getters->name);
		getters++;
	}
	printf ("\n");
}

/**
 * \brief Export (print record content as CSV) expired flow records
 *
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \param[in] record Flow record from input plugin(s)
 * \return
 *   - zero on success,
 *   - nonzero on error
 */
int plugin_export_export(void *plugin_private, flow_record_t *record)
{
	bacnet_export_csv_private_t *p = (bacnet_export_csv_private_t*)plugin_private;
	flow_record_getter_t *tmp = p->getters;
	unsigned char block[1024];
	int i, j, len, to_network_byte_order;
	char addr[ADDR_BUF_SIZE];

	char datestr[64];
	time_t t = time(NULL);
	struct tm *now;

	PRINT_SQL_INIT;

	datestr[0] = '\0';

	/*
	 * Print the first line with headers of printed columns
	 *
	if (p->counter == 0) {
		print_headers(tmp);
	}
	*/

	p->counter++;
	for (i = 0; selected_getters[i].id != GETTER_MAX; i++) {
		if ((tmp = getter_by_name (p->getters, selected_getters[i].name)) == NULL) {
			continue;
		}
		len = tmp->current_length(tmp->self,record);
		if ((selected_getters[i].id == L3_IPV6_ADDR_SRC) ||
			(selected_getters[i].id == L3_IPV6_ADDR_DST) ||
			(selected_getters[i].id == L3_IPV4_ADDR_SRC) ||
			(selected_getters[i].id == L3_IPV4_ADDR_DST)) {
			to_network_byte_order = 1;
		} else {
			to_network_byte_order = 0;
		}
		if(!tmp->valid(tmp->self,record)){
			PRINT_SQL(",NULL");
			continue;
		}
		tmp->filler (tmp->self, record, block, len, to_network_byte_order);

		/* process data with appropriate getter functions */
		switch (selected_getters[i].id) {
		case FLOW_START_USEC:
			/* handle partitioning if enabled */
			if (p->partition != TU_NO) {
				t = (time_t) (*(uint64_t*)&(block[0]))/1000000;
				now = localtime (&t);
				switch (p->partition) {
				case TU_YEAR:
					snprintf (datestr, 63, "_%i", now->tm_year + 1900);
					break;
				case TU_MONTH:
					snprintf (datestr, 63, "_%i%02i", now->tm_year + 1900, now->tm_mon + 1);
					break;
				case TU_DAY:
					snprintf (datestr, 63, "_%i%02i%02i", now->tm_year + 1900, now->tm_mon + 1, now->tm_mday);
					break;
				case TU_HOUR:
					snprintf (datestr, 63, "_%i%02i%02i%02i", now->tm_year + 1900, now->tm_mon + 1, now->tm_mday, now->tm_hour);
					break;
				case TU_MINUTE:
					snprintf (datestr, 63, "_%i%02i%02i%02i%02i", now->tm_year + 1900, now->tm_mon + 1, now->tm_mday, now->tm_hour, now->tm_min);
					break;
				default:
					/* TU_NO - not possible to be here */
					break;
				}
			}
			PRINT_SQL("INSERT INTO bacnet%s VALUES (to_timestamp('%lu.%06lu'),", datestr, (*(uint64_t*)&(block[0]))/1000000, (*(uint64_t*)&(block[0]))%1000000);
			//PRINT_SQL("INSERT INTO bacnet VALUES ('%lu.%06lu',", (*(uint64_t*)&(block[0]))/1000000, (*(uint64_t*)&(block[0]))%1000000);
			break;
		case FLOW_END_USEC:
			PRINT_SQL("to_timestamp('%lu.%06lu')", (*(uint64_t*)&(block[0]))/1000000, (*(uint64_t*)&(block[0]))%1000000);
			//PRINT_SQL("'%lu.%06lu'", (*(uint64_t*)&(block[0]))/1000000, (*(uint64_t*)&(block[0]))%1000000);
			break;
		case L2_SRC_MAC:
		case L2_DST_MAC:
		case BACNET_SADR:
		case BACNET_DADR:
			PRINT_SQL(",'");
			/* (MAC) address */
			if (len==1) {
				PRINT_SQL("00:00:00:00:00:%02x", block[0]);	/* Source MAC Layer Address Length: 1 e.g. SADR: 11 */
			}
			else {
				PRINT_SQL("%02x", block[0]);
				for (j = 1; j < len; j++) {
					PRINT_SQL(":%02x", block[j]);
				}
			}
			PRINT_SQL("'");
			break;
		case BACNET_CONTROL:
			PRINT_SQL(",'%u'", block[0]);
			break;
		case BACNET_HOP_COUNT:
		case BACNET_MESSAGE_TYPE:
		case L3_PROTO:
		case L4_PROTO:
		case L4_TCP_FLAGS:
		case L3_IPV4_TOS:
			/* one-byte numbers */
			PRINT_SQL(",'%u'", block[0]);
			break;
		case L2_VLAN_0:
		case L2_VLAN_1:
		case L2_VLAN_2:
		case L2_VLAN_3:
		case BACNET_VENDOR_ID:
		case ETHERTYPE:
		case BACNET_SNET:
		case BACNET_DNET:
		case L4_PORT_SRC:
		case L4_PORT_DST:
		case L4_ICMP_TYPE_CODE:
			/* two-byte numbers */
			PRINT_SQL(",'%u'", (*(uint16_t*)block));
			break;
		case PACKETS:
		case BYTES:
			/* 32b number */
			PRINT_SQL(",'%u'", (*(uint32_t*)block));
			break;
		case L3_IPV6_ADDR_SRC:
		case L3_IPV6_ADDR_DST:
			PRINT_SQL(",'");
			inet_ntop (AF_INET6, block, addr, ADDR_BUF_SIZE);
			PRINT_SQL("%s/128'", addr);
			break;
		case L3_IPV4_ADDR_SRC:
		case L3_IPV4_ADDR_DST:
			PRINT_SQL(",'");
			inet_ntop (AF_INET, block, addr, ADDR_BUF_SIZE);
			PRINT_SQL("%s/32'", addr);
			break;
		default:
			break;
		}
	}

	PRINT_SQL(");\n");
	PRINT_SQL_FLUSH;	/* flush stdout to avoid data corruption during writing to text file */
	return(0);
}

/**
 * \brief Flush data.
 *
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \return
 *   - zero on success,
 *   - nonzero on error
 */
int plugin_export_flush(void *plugin_private)
{
	PRINT("Output plugin flush\n");
	fflush(stdout);
	return(0);
}

/**@}*/

