/**
 * \file filter-bacnet.c
 * \brief BACnet Monitoring Plugin - BACnet filter plugin.
 * \copyright Masaryk University, Institute of Computer Science
 * \date 2011
 * \author Pavel CELEDA  <celeda@ics.muni.cz>
 * \author Radek KREJCI  <radek.krejci@mail.muni.cz>
 *
 * Copyright (C) 2011 Masaryk University, Institute of Computer Science
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Masaryk University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: filter-bacnet.c 175 2012-01-10 12:26:31Z 98863 $
 */
static const char rcsid[] __attribute__((used)) ="$Id: filter-bacnet.c 175 2012-01-10 12:26:31Z 98863 $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/if_ether.h>
#include <flowmonexp/plugin_filter.h>

#include "../input-plugins/protocols.h"

/**
 * \defgroup bacnetfilter BACnet filter plugin
 *
 * \brief Plugin filters flow record get by input plugin according to its
 * EtherType value (if any such value is present).
 *
 * Default behavior of the plugin is to allow only BACnet, ARP, ISMP, LLDP and
 * SLOW (802.3) packets. Any other type of network traffic is dropped. This can
 * be inverted by a 'deny' plugin parameter:
 * \verbatim $ flowmonexp -X input-bacnet.so -X filter-bacnet.so -I input-bacnet -F filter-bacnet:deny -E netflow-v9:host=localhost\endverbatim
 *
 * @{
 */

/**
 * Plugin version
 */
#define VERSION "1.0"

/**
 * Message printer
 */
#define PRINT(format,args...)
//#define PRINT(format,args...) fprintf(stderr,format,##args)

/**
 * Error message printer
 */
#define PRINTERR(format,args...) fprintf(stderr,format,##args)

/**
 * Warning message printer
 */
#define PRINTWARN(format,args...) fprintf(stderr,format,##args)

/**
 * \brief Plugin's private structure with information passed from init to the
 * calls of PLUGIN_FILTER_* functions.
 */
typedef struct
{
	flow_record_getter_t *ethertype_getter; /**< \brief Getter for EtherType value */
	int not; 				/**< \brief logic for filter behavior, values 0 for allow (default) and 1 for deny */
}bacnet_filter_private_t;

/**
 * Set type of the flowmonexp's plugin
 */
unsigned int __attribute__((used))plugin_type = (PLUGIN_TYPE_FILTER);

/**
 * \brief Plugin description structure for internal use in flowmonexp
 */
static plugin_desc_t plugin_desc = {
	"filter-bacnet", 	/**< Plugin name */
	"Version " VERSION
	"\nPlugin pass only BACnet and some other non-IP data (ARP). The behavior can be inverted by \'deny\' option.", /**< Plugin help string */
	0, 	/**< Requested data size after standard flow record (MUST BE CONSTANT) */
	0 	/**< Plugin must have packet input, valid only for process plugin */
};

/**
 * \brief API for flowmonexp to get plugin description structure.
 * \return Plugin description structure
 */
plugin_desc_t *plugin_filter_desc()
{
	return(&plugin_desc);
}

/**
 * \brief Initialize all needed structures/devices/files etc. The function is
 * called once as the first part of the plugin.
 *
 * \param[in] params Plugin parameters as string from command line
 * \param[in] getter_list List of available getters for a complete flow record
 * \return Created and initialized plugin internal structure
 *         (filter_bacnet_private_t)
 */
void *plugin_filter_init(char *params, flow_record_getter_t **getter_list)
{
	bacnet_filter_private_t *retval;
	flow_record_getter_t *tmp;

	PRINT ("plugin filter-bacnet init start\n");
	retval = malloc(sizeof(bacnet_filter_private_t));
	if(!retval){
		return(NULL);
	}

	retval->ethertype_getter = NULL;
	retval->not = 0;

	if (params != NULL) {
		if (strlen (params) == 0) {
			PRINTERR ("Invalid parameters\n");
			exit (-1);
		}

		if (!strcmp ("allow", params)) {
			retval->not = 0;
		} else if (!strcmp ("deny", params)) {
			retval->not = 1;
		} else {
			/* unknown option */
			PRINTERR ("Invalid parameters\n");
			exit (-1);
		}
	}

	tmp = getter_by_name(getter_list[0], "ETHERTYPE");
	if(tmp == NULL){
		PRINTERR ("Unable to find Ethertype getter\n");
		exit(-1);
	}
	getter_copy_to(&(retval->ethertype_getter), tmp);

	PRINT ("plugin filter-bacnet init end\n");

	return(retval);
}

/**
 * \brief Filtering function. By default (allow behavior), filter passes only
 * BACnet, ARP, ISMP, LLDP and SLOW (802.3) protocols. Other traffic (like IP)
 * is discarded. Deny parameter of the plugin inverts the behavior.
 *
 * \param[in] plugin_private Plugin's private structure from init with all necessary information
 * \param[in] record Flow record from input plugin(s)
 * \return
 *   - FLOW_FILTER_DROP to discard flow record,
 *   - FLOW_FILTER_PASS to allow further processing/exporting of the flow record
 */
int plugin_filter_filter(void *plugin_private, flow_record_t *record)
{
	bacnet_filter_private_t *p = (bacnet_filter_private_t*)plugin_private;
	uint16_t ethertype;

	if (p->ethertype_getter && p->ethertype_getter->name){
		if(p->ethertype_getter->valid(p->ethertype_getter->self,record)){
			if (p->ethertype_getter->current_length(p->ethertype_getter->self, record) != 2) {
				return (FLOW_FILTER_DROP);
			}

			/* get ethertype */
			p->ethertype_getter->filler(p->ethertype_getter->self, record, &ethertype, 2, 0);

			switch (ethertype) {
			case BACNET_ETH_TYPE:
			case BACNET_IP_TYPE:
			case ETH_P_ARP:
			case ETH_P_ISMP:	/* InterSwitch Message Protocol */
			case ETH_P_SLOW:	/* Slow Protocols (IEEE 802.3) */
			case ETH_P_LLDP:	/* Link Layer Discovery Protocol */
				if(p->not){
					return(FLOW_FILTER_DROP);
				} else {
					return(FLOW_FILTER_PASS);
				}
				break;
			default:
				if(p->not){
					return(FLOW_FILTER_PASS);
				} else {
					return(FLOW_FILTER_DROP);
				}
				break;
			}
		}
	}

	if(p->not){
		return(FLOW_FILTER_PASS);
	} else {
		return(FLOW_FILTER_DROP);
	}
}

/** @}*/

